 /**
 *  \file    KsCmakeDef.hpp
 *  \brief   This File is generated by CMAKE
 */

/* ! -- Do Not Hand Edit - This File is generated by CMAKE -- ! */

#ifndef _KS_CONFIG_H
#define _KS_CONFIG_H

/** KernelShark Version number. */
#define KS_VERSION_STRING "2.4.0"

/** KernelShark Application name. */
#define KS_APP_NAME "kernelshark"

/** KernelShark installation prefix path. */
#define _INSTALL_PREFIX "/usr"

/** KernelShark plugins installation prefix path. */
#define KS_PLUGIN_INSTALL_PREFIX "/usr/lib/kernelshark/plugins"

/** Location of the trace-cmd executable. */
#define TRACECMD_EXECUTABLE "/usr/bin/trace-cmd"

/** "pkexec" executable. */
/* #undef DO_AS_ROOT */

/** GLUT has been found. */
#define GLUT_FOUND

/** Truetype font file. */
#define TT_FONT_FILE "/usr/share/fonts/TTF/FiraSans-Regular.ttf"

/** Location of the KernelShark tests. */
#define KS_TEST_DIR "/build/kernelshark/src/kernel-shark-kernelshark-v2.4.0/tests"

/** Semicolon-separated list of plugin names. */
#define KS_BUILTIN_PLUGINS "sched_events;event_field_plot;latency_plot;kvm_combo;missed_events"

#endif // _KS_CONFIG_H
